/* dhcp4_nic.c
 *
 * Network Interface Configurator for the ISC DHCP IPv4 client library.
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <dhcp4_nic.h>
#include <dhcp4client.h>
#include <isc-dhcp/dhcpd.h>

struct dhcpv4_control_s
{
/*{ this structure MUST be the same as in libdhcp_control.h for DHCP clients */
    LIBDHCP_Callback    callback;    	/* the DHCP clients' main loop calls this on state changes */
    uint16_t            capability;     /* LIBDHCP_Capability bits to enable                       */
    uint8_t             finished;       /* set to one to make clients exit their main loop         */
    uint8_t             decline;        /* set to one to decline the lease (DHCPv4 only)           */
    time_t              timeout;        /* (timeout+now) == time after which clients MUST return   */
    time_t              now;            /* clients set this to time(0) on entering main loop       */
    void               *arg;            /* user data pointer                                       */
    LIBDHCP_Error_Handler eh;
/*} end of fields known to DHCP clients */
    uint8_t             log_level;      /* maximum log level (LOG_FATAL excluded)                  */    
    NLH_t               nh;
/* struct DHCPv4_nic { */
    NIC_t               nic;
    IPaddr_list_t       addr_list;
    IProute_list_t      route_list;
    IPaddr_list_t       dns_list;
    char               *search_list;
    char               *host_name;
    DHCPv4_lease       *lease;
/*                   } */
    IPaddr_t           ip4;
    char               *if_name;
    char              **argv;
    int                 argc;
};

static void
dhc4_log( DHCPv4_control *ctl, int priority, char *fmt, ...)
{
    va_list va;
    if((ctl == 0)|| ((ctl->eh == 0) || (ctl->log_level < priority)))
	return;
    va_start(va, fmt);
    ctl->eh((LIBDHCP_Control*)ctl, priority, fmt, va);
    va_end(va);
}

DHCPv4_control*
dhcpv4_control
( 
    NLH_t                  nic_handle,
    char                  *eth_if_name,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level,
    ... /* extra dhclient cli args */
)
{
    DHCPv4_control *dhc = calloc( 1, sizeof( DHCPv4_control ) ); 
    va_list va;
    char **argv, *a[32]={strdup("dhclient")}, *p;
    int argc;
    uint8_t d_needed=1, if_needed=1;

    if( dhc == 0L )
	return 0;

    dhc->nh = nic_handle;

    if( (dhc->nic = nic_by_name(dhc->nh, eth_if_name)) == 0L )
	return 0;

    dhc->capability = dhc_cap;
    dhc->callback = dhcp4_nic_callback;
    dhc->timeout = timeout;
    dhc->arg = 0;
    dhc->eh = error_handler;
    dhc->log_level = log_level;
    dhc->if_name = strdup(eth_if_name);

    va_start(va, log_level);
    
    for(argc=1, argv=&(a[1]); (argc < 31) && (p = va_arg(va, char *)); argc++, argv++)
    {
	*argv = strdup(p);
	if( strcmp(p,"-d") == 0)
	    d_needed = 0;
	if( strcmp(p,eth_if_name) == 0)
	    if_needed = 0;	
    }
    va_end(va);

    if ( d_needed )
    {
	if ( argc > 31 )
	{
	    dhcpv4_control_free(dhc);
	    return 0L;
	}
	a[argc++] = strdup("-d");
    }

    if ( if_needed )
    {
	if ( argc > 31 )
	{
	    dhcpv4_control_free(dhc);
	    return 0L;
	}
	a[argc++] = strdup(eth_if_name);
    }

    a[argc] = 0L;
    argv = calloc(argc, sizeof(char*));
    if( argv == 0 )
    {
	free(dhc);
	return 0L;
    }
    dhc->argc = argc;
    
    memcpy(argv,&a[0],argc * sizeof(char*));
    dhc->argv = argv;
    STAILQ_INIT( &(dhc->addr_list) );
    STAILQ_INIT( &(dhc->route_list) );
    STAILQ_INIT( &(dhc->dns_list) );    
    return dhc;
}

void dhcpv4_control_free( DHCPv4_control *dhc )
{
    if(dhc->lease)
    {
	dhcpv4_lease_free(dhc->lease);
	dhc->lease = 0;
    }
    if(dhc->if_name)
    {
	free(dhc->if_name);
	dhc->if_name = 0;
    }
    if( dhc->argv )
    {
	char **p;
	for(p=dhc->argv; *p; p++)
	    free(p);
	free(dhc->argv);
	dhc->argv = 0;
    }
    if( !STAILQ_EMPTY( &(dhc->addr_list) ) )
	nic_address_list_free(&(dhc->addr_list));
    if( !STAILQ_EMPTY( &(dhc->dns_list) ) )
	nic_address_list_free(&(dhc->dns_list));
    if( !STAILQ_EMPTY( &(dhc->route_list) ) )
	nic_route_list_free(&(dhc->route_list));
       
    free(dhc);
}

extern char **environ;

DHCPv4_nic *do_dhcpv4( DHCPv4_control *dh4c )
{
    dhcpv4_client((LIBDHCP_Control*)dh4c, dh4c->argc, dh4c->argv, environ);
    return (DHCPv4_nic*)&(dh4c->nic);
}

NIC_Res_t dhcpv4_nic(NLH_t nh, DHCPv4_nic *nic )
{
    return 
	nic_configure
        (
	    nh,
	    nic->nic,
	    &(nic->address_list),
	    &(nic->route_list),
	    &(nic->dns_list),
	    nic->search_list,
	    nic->host_name
	);
}

int dhcp4_nic_callback
(   LIBDHCP_Control *cp,
    DHCP_State state, 
    void *arg
)
{
    DHCPv4_control *control = (DHCPv4_control *) cp;
    dhc4_log
    (   control, LOG_DEBUG,
	"DHCPv4 %s  - state: %s if: %s arg: %p\n", 	
	control->if_name,
	libdhcp_state_string(state,0),
	control->arg, arg
    );
    switch( state )
    {
    case DHC_TIMEDOUT:
	dhc4_log
	(   control, LOG_INFO,
	    "DHCPv4 %s  - TIMED OUT.",
	    control->if_name
	);
	control -> finished = 1;
	break;
	
    case DHC4_PREINIT:
    {	/* bring interface UP */
	NIC_t nic = control->nic;
	uint32_t iff = nic_get_flags(nic);
	dhc4_log
	(   control, LOG_INFO,
	    "DHCPv4 %s  - bringing link UP - %x %d",
	    control->if_name,
	    iff, iff & (IFF_UP | IFF_RUNNING) 
	);
	if ( ( iff & (IFF_UP | IFF_RUNNING) ) != (IFF_UP | IFF_RUNNING) ) 
	    if ( nic_set_flags(nic, iff | IFF_UP | IFF_RUNNING ) != NIC_OK )
	    {
		dhc4_log
		(   control, LOG_ERR,
		    "DHCPv4 %s - bringing link UP failed.",
		    control->if_name
		);
		return 1;
	    }	
    }   break;

    case DHC4_BOUND:
    case DHC4_REBOOT:
    case DHC4_RENEW:
    case DHC4_REBIND:
    case DHC4_TIMEOUT: /* Cannot currently contact server, but we have an active lease */
    {	 
	/* process new lease */
	DHCPv4_lease *lease = dhcpv4_lease( arg );
	if ( lease )
	{
	    char buf[32];
	    struct sockaddr sa;
	    dhc4_log
	    (   control, LOG_INFO,
		"DHCPv4 %s - obtained lease %s",
		control->if_name,
		inet_ntop(AF_INET, &(lease->address), buf, 32)
	    );
	    control->lease = lease;
	    memset(&sa, '\0', sizeof(struct sockaddr));
	    memcpy(&(((struct sockaddr_in*)&sa)->sin_addr), &(lease->address.s_addr),4);
	    sa.sa_family = AF_INET;
	    control->ip4 = nic_sa_addr( control->nh, &sa, sizeof(struct in_addr));
	    IPaddr_list_node_t *n = calloc(1,sizeof(IPaddr_list_node_t));
	    n->addr = control->ip4;
	    STAILQ_INSERT_TAIL( &(control->addr_list), n, link );

	    dhcpv4_process_options(lease, dhcp4_nic_option_handler, control);
	}
	control -> finished = 1;
	break;
    }

    case DHC4_RELEASE:
    case DHC4_EXPIRE:
    case DHC4_FAIL:
    case DHC4_STOP:
	/* unconfigure existing lease */
	/* XXX - should do something here!*/
	control -> finished = 1;
	break;

    default:
	dhc4_log
	(   control, LOG_ERR,
	    "DHCPv4 %s - entered unhandled state.",
	    control->if_name
	);
    }
    return 0;
}

void dhcp4_nic_option_handler( DHCPv4_option *option, void *arg )
{
    DHCPv4_control *control = arg;
    char buf[32];
    
    if ( option -> unicode == DHC_DHCP_Universe )
    {
	switch( option -> code )
	{
	case DHCO_SUBNET_MASK:
	{
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option subnet-mask: %s", 
		control->if_name,
		inet_ntop( AF_INET, (struct in_addr*)&(option->value[0]), &(buf[0]), sizeof(buf))
	    );
	    uint32_t sm=0;
	    uint32_t  i=32, b=1;
	    memcpy(&sm, &(option->value[0]), sizeof(uint32_t));
	    sm = ntohl(sm);
	    for(; i && ((sm & b) != b); i-=1, b <<= 1);
	    nic_addr_set_prefix( control->ip4, i);
	    IPaddr_t ip4_broadcast = nic_ipv4_broadcast_addr( control->ip4 );
	    nic_addr_set_broadcast( control->ip4, ip4_broadcast );
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option subnet-mask - prefix_len: %d broadcast: %s", 
		control->if_name, i, nic_addr_to_text(ip4_broadcast, buf, 32)		
	    );
	}
	break;

	case DHCO_ROUTERS:
	{
	    int i;
	    struct in_addr *routers = (struct in_addr*) &(option->value[0]);
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option routers:", 
		control->if_name
	    );
	
	    for( i = 0; i < option->n_elements; i++, routers++ )
	    {
		dhc4_log
	        (   control, LOG_DEBUG, "DHCPv4 %s - option routers default gateway %d: %s", 
		    control->if_name, i,
		    inet_ntop( AF_INET, routers, &(buf[0]), sizeof(buf))
		);
		struct sockaddr_in gw;
		memset(&gw,'\0',sizeof(struct in_addr));
		memcpy(&(gw.sin_addr.s_addr), routers, sizeof(struct in_addr));
		gw.sin_family = AF_INET;
		IProute_t route =
		    nic_route_new
		    (   control->nh,
			nic_get_index(control->nic),
			0L,0,                      /* DEFAULT ROUTE ! */
			(struct sockaddr*)&gw,     /* gateway */
			-1,                        /* default scope: global */
			(i > 0) ? (int8_t)i : -1,  /* priority: increasing if more than one */
			-1,                        /* table: default (local) */
			-1/*no iif*/, 0L,0/*no src*/		    
		    );		
		IProute_list_node_t *n = calloc(1,sizeof(IProute_list_node_t));
		n->route = route;
		STAILQ_INSERT_TAIL(&(control->route_list),n,link);
	    }
	}
	break;	
	case DHCO_STATIC_ROUTES:
	{
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option static-routes:", 
		control->if_name
	    );

	    char buf2[32];
	    int i;
	    struct { struct in_addr dst, gw; } *routes = (void*)&(option->value[0]);
	    IProute_t route;

	    for( i = 0; i < option->n_elements; i++, routes++ )
	    {
		dhc4_log
		(   control, LOG_DEBUG, "DHCPv4 %s - option static-routes - route %d: %s via %s",
		    control->if_name, 
		    inet_ntop( AF_INET, &(routes->dst), &(buf[0]), sizeof(buf)),
		    inet_ntop( AF_INET, &(routes->gw), &(buf2[0]), sizeof(buf))
		);

		struct sockaddr_in dst;
		memset(&dst,'\0',sizeof(struct in_addr));
		memcpy(&(dst.sin_addr.s_addr), &(routes->dst), sizeof(struct in_addr));
		dst.sin_family = AF_INET;
		uint32_t dip=0;
		memcpy(&dip, &(routes->dst), sizeof(uint32_t));
		dip = ntohl(dip);
		uint32_t  dst_len=32, b=1;
		for(; dst_len && ((dip & b) != b); dst_len--, b <<= 1);

		struct sockaddr_in gw;
		memset(&gw,'\0',sizeof(struct in_addr));
		memcpy(&(gw.sin_addr.s_addr), &(routes->gw), sizeof(struct in_addr));
		gw.sin_family = AF_INET;

		route = 
		    nic_route_new
		    (   control->nh,
			nic_get_index(control->nic),
			(struct sockaddr*)&dst, dst_len,
			(struct sockaddr*)&gw,     /* gateway */
			-1,                        /* default scope: global */
			(i > 0) ? (int8_t)i : -1,  /* priority: increasing if more than one */
			-1,                        /* table: default (local) */
			-1/*no iif*/, 0L,0/*no src*/		    
		    );			       
		IProute_list_node_t *n = calloc(1,sizeof(IProute_list_node_t));
		n->route = route;
		STAILQ_INSERT_TAIL(&(control->route_list),n,link);
	    }
	}
	break;

	case DHCO_DOMAIN_NAME:
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option domain-name: %s",
		control->if_name, (char*)(option->value)
	    );
	    control->search_list = strdup( (char*)option->value );
	    break;

	case DHCO_HOST_NAME:
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option host-name: %s",
		control->if_name, (char*)(option->value)
	    );
	    control->host_name = strdup( (char*)option->value );
	    break;

	case DHCO_DOMAIN_NAME_SERVERS:
	{
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option domain-name-servers:",
		control->if_name
	    );
	    int i;
	    struct in_addr *dns = (struct in_addr*) &(option->value[0]);
	    for( i = 0; i < option->n_elements; i++, dns++ )
	    {
		dhc4_log
	        (   control, LOG_DEBUG, "DHCPv4 %s - domain-name-server %d: %s",
		    control->if_name, i,
		    inet_ntop(AF_INET, dns, buf, sizeof(buf))
		);
		struct sockaddr_in dns_sa;
		memset(&dns_sa,'\0',sizeof(struct sockaddr_in));
		dns_sa.sin_addr.s_addr = dns->s_addr;
		dns_sa.sin_family = AF_INET;
		IPaddr_t dnsIP=nic_sa_addr(control->nh, (struct sockaddr*)&dns_sa, sizeof(struct in_addr));
		IPaddr_list_node_t *n = calloc(1,sizeof(IPaddr_list_node_t));
		n->addr = dnsIP;
		STAILQ_INSERT_TAIL(&(control->dns_list), n, link);
	    }
	}
	break;
	default:
	    /* XXX: many more options to do something about! */
	    break;
	}
    }
}
