/* dhcp6_lease.c
 *
 *  Representation of DHCPv6 lease options for libdhcp
 *
 *  Copyright(C) Jason Vas Dias<jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <resolv.h>
#include <string.h>
#include <time.h>
#include <stdio.h>

/* dhcpv6 includes: */

#include <sys/queue.h>
/* #include <queue.h> 
 * dhcpv6 uses an older copy of BSD queue.h, which 
 * is happily compatible with the new version in glibc
 */
#define  LIBDHCP
#include <dhcpv6/dhcp6.h>
#include <dhcpv6/config.h>
#include <dhcpv6/lease.h>
#undef malloc
#undef realloc
#undef calloc
#undef free
#include <malloc.h>
#include <dhcp6_lease.h>


DHCPv6_lease *dhcpv6_lease( struct dhcp6_optinfo *optinfo )
{
    if ( TAILQ_EMPTY( &(client6_iaidaddr.lease_list) ) )
	return 0L;

    DHCPv6_lease *lease = calloc(1, sizeof(DHCPv6_lease));
				 
    if( lease == 0L )
	return 0;
    
    STAILQ_INIT(&(lease->addresses));
    STAILQ_INIT(&(lease->dns));

    struct dhcp6_lease *l;
    for (l = TAILQ_FIRST( &(client6_iaidaddr.lease_list) ); l; l = TAILQ_NEXT(l,link))
    {
	if(  ( l->state == DHCP6_ACTIVE )
	  && ( l->lease_addr.type != DHCPv6_DELEGATED_PREFIX )
	  && (  ( l->lease_addr.status_code  == DH6OPT_STCODE_SUCCESS ) 
	      ||( l->lease_addr.status_code  == DH6OPT_STCODE_UNDEFINE )
	     )
	  ) 
	{
	    DHCPv6_lease_address *addr = calloc(1, sizeof( DHCPv6_lease_address ) );
	    if( addr == 0 )
		goto dhcpv6_lease_failure;
	    
	    memcpy(&(addr->address), &(l->lease_addr.addr), sizeof(struct in6_addr));
	    
	    addr -> prefix_length =  l->lease_addr.plen;
	    addr -> type =  l->lease_addr.type;
	    addr -> status = l->lease_addr.status_code;
    	    addr -> time_acquired = l->start_date;
	    addr -> prefer_lifetime = l->lease_addr.preferlifetime;
	    addr -> valid_lifetime = l->lease_addr.validlifetime;
	    
	    STAILQ_INSERT_TAIL( &(lease->addresses), addr, link );
	}
    }

    lease->rebind_time = client6_iaidaddr.client6_info.iaidinfo.rebindtime;
    lease->renew_time =  client6_iaidaddr.client6_info.iaidinfo.renewtime;
    lease->iaid = client6_iaidaddr.client6_info.iaidinfo.iaid;
    lease->client_duid = malloc(client6_iaidaddr.client6_info.clientid.duid_len );
    if( lease->client_duid  == 0L )
	goto dhcpv6_lease_failure;
    memcpy(lease->client_duid,  client6_iaidaddr.client6_info.clientid.duid_id,
	                        client6_iaidaddr.client6_info.clientid.duid_len);
    lease->server_duid = malloc(client6_iaidaddr.client6_info.serverid.duid_len );
    if( lease->server_duid  == 0L )
	goto dhcpv6_lease_failure;
    memcpy(lease->server_duid,  client6_iaidaddr.client6_info.serverid.duid_id,
	                        client6_iaidaddr.client6_info.serverid.duid_len);    
    memcpy(&(lease->server_address), &(optinfo->server_addr), sizeof(struct in6_addr) );
    if( (lease->if_name = strdup( client6_iaidaddr.ifp->ifname )) == 0L )
 	goto dhcpv6_lease_failure;    
    lease->if_index = if_nametoindex( lease->if_name );
    
    if ( ! TAILQ_EMPTY( &(optinfo -> dns_list.addrlist) ) )
    {
	struct dhcp6_listval *v;	
	IPv6_address *dns;
	int i=0;
	for (v = TAILQ_FIRST(&(optinfo->dns_list.addrlist)); v && (i < MAXNS); v = TAILQ_NEXT(v, link), i++) 
	{
	    dns = calloc(1, sizeof(IPv6_address));
	    if( dns == 0L )
		goto dhcpv6_lease_failure;
	    memcpy(&(dns->address), &(v->val_addr6), sizeof(struct in6_addr));
	    STAILQ_INSERT_TAIL(&(lease->dns), dns, link);
	} 
    }
    if ( optinfo -> dns_list.domainlist != 0 )
    {
	int i=0, len;
	struct domain_list *dlist;
	char search_list[256], *p=search_list;
	for (dlist = optinfo->dns_list.domainlist; dlist && (i < MAXDNSRCH); dlist = dlist->next, i++)
	{
	    len = strlen(dlist->name);
	    if ( (len + 1 + (p - search_list)) >= sizeof(search_list) )
		break;
	    p += sprintf(p,"%s ", dlist->name);
	}
	if( p - search_list )
	    if( ( lease->search_list = strdup(search_list) ) == 0L )
		goto dhcpv6_lease_failure;
    }
    
    return lease;

    dhcpv6_lease_failure:
    dhcpv6_lease_free( lease );
    return 0L;
}

void dhcpv6_lease_free( DHCPv6_lease *lease )
{
    if( !STAILQ_EMPTY(&(lease->addresses)) )
    {
	DHCPv6_lease_address *addr = STAILQ_FIRST( &(lease->addresses) ), *next;
	do
	{
	    next = STAILQ_NEXT(addr, link);
	    free( addr );
	    addr = next;
	} while ( addr );	
    }
    if ( !STAILQ_EMPTY( &(lease->dns) ) )
    {
	IPv6_address *addr = STAILQ_FIRST( &(lease->dns) ), *next;
	do
	{
	    next = STAILQ_NEXT(addr, link);
	    free( addr );
	    addr = next;
	} while ( addr );
    }
    if ( lease->search_list )
	free(lease->search_list);
    if ( lease->if_name )
	free(lease->if_name);
    if ( lease->client_duid )
	free(lease->client_duid);
    if ( lease->server_duid )
	free(lease->server_duid);
}
