/* libdhcp.h
 *
 *  API for libdhcp, a minimal interface to the ISC dhcp IPv4 client,
 *  and to the dhcpv6 DHCPv6 client libraries, libdhcp4client, and 
 *  libdhcp6client .
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#ifndef LIBDHCP_H
#define LIBDHCP_H

#include <sys/types.h>
#include <sys/syslog.h>
#include <stdint.h>
#include <stdarg.h>

#define LIBDHCP 1

enum dhcp_state_e;
struct libdhcp_control_s;

typedef 
int ( *LIBDHCP_Callback ) ( struct libdhcp_control_s *control, enum dhcp_state_e, void* );
    /* The DHCP clients will call the users' callback on important state change events,
     * with the second arg set to the client DHCP_State, and the third arg set to
     * a client specific pointer as described below.
     */
#define LOG_FATAL 8
typedef
int ( *LIBDHCP_Error_Handler )
    ( struct libdhcp_control_s *ctl,
      int priority,  /* ala syslog(3): LOG_EMERG=0 - LOG_DEBUG=7 (+ LOG_FATAL=8 : finished -> 1)   */
      const char *fmt,
      va_list ap
    );

typedef 
struct libdhcp_control_s
{
/*{ this structure MUST be the same as in libdhcp_control.h for DHCP clients */
    LIBDHCP_Callback    callback;    	/* the DHCP clients' main loop calls this on state changes */
    uint16_t            capability;     /* LIBDHCP_Capability bits to enable                       */
    uint8_t             finished;       /* set to one to make clients exit their main loop         */
    uint8_t             decline;        /* set to one to decline the lease (DHCPv4 only)           */
    time_t              timeout;        /* (timeout+now) == time after which clients MUST return   */
    time_t              now;            /* clients set this to time(0) on entering main loop       */
    void               *arg;            /* user data pointer                                       */
    LIBDHCP_Error_Handler eh;
/*} end of fields known to DHCP clients */
    uint8_t             log_level;      /* maximum log level (LOG_FATAL excluded)                  */    
} LIBDHCP_Control;

typedef enum libdhcp_capability_e
{/* DHCP client "capabilities" */ 
    DHCP_USE_LEASE_DATABASE   = 1,  	/* use / do not use persistent lease database files */
    DHCP_USE_PID_FILE         = 2,  	/* use / do not use pid file                        */
 /*
  * DHCPv6 supports these capabilities in process, 
  * while the DHCPv4 client will fork and exec the dhclient-script to implement them if these
  * bits are set - otherwise, if no bits are set, the callback is called and the script is 
  * not run.
  */
    DHCP_CONFIGURE_INTERFACES = 4,  	/* configure interfaces UP/DOWN as required         */
    DHCP_CONFIGURE_ADDRESSES  = 8,  	/* configure interface addresses as required        */
    DHCP_CONFIGURE_ROUTES     =16,  	/* configure routes as required                     */
    DHCP_CONFIGURE_RESOLVER   =32,  	/* configure resolv.conf as required                */
    /* DHCPv6 only: */
    DHCP_CONFIGURE_RADVD      =64,  	/* configure radvd.conf & restart radvd as required */
} LIBDHCP_Capability;

/* WARNING: use of DHCP_USE_LEASE_DATABASE with DHCPv6 when invoking dhcpv6_client multiple
 * times in the same process will lead to memory leaks in the flex scanner.
 * This problem is being worked on, but is expected most users of the dhcp6client library
 * will be invoking the client only once / not using DHCP_USE_LEASE_DATABASE .
 * DHCPv4 has no memory leaks.
 */

extern LIBDHCP_Control*
libdhcp_control_new
(   LIBDHCP_Callback   	   callback,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,
    void                  *arg,
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level
);  /* creates new LIBDHCP_Control handle */


typedef int (*DHCP_Client) ( LIBDHCP_Control*, int argc, char **argv, char **envp );

extern int
libdhcp_call_client( LIBDHCP_Control *,  DHCP_Client, ...);
    /* calls the dhcp[46]_client function with arguments in variable argument list.
     * The last argument in the argument list MUST be 0 .
     */

typedef enum dhcp_state_e
{

    /* DHCPv4 client states - third callback arg will be a 'struct client_state *'    */
    DHC4_NBI,         		/* failed: no broadcast interfaces found              */
    DHC4_PREINIT, 		/* configuration started - bring the interface "UP"   */
    DHC4_BOUND, 		/* lease obtained                                     */
    DHC4_RENEW, 		/* lease renewed                                      */
    DHC4_REBOOT,	        /* have valid lease, but now obtained a different one */
    DHC4_REBIND, 		/* new, different lease                               */
    DHC4_STOP,  		/* remove old lease                                   */
    DHC4_MEDIUM, 		/* media selection begun                              */
    DHC4_TIMEOUT, 		/* timed out contacting DHCP server                   */
    DHC4_FAIL, 			/* all attempts to contact server timed out, sleeping */
    DHC4_EXPIRE, 		/* lease has expired, renewing                        */
    DHC4_RELEASE, 		/* releasing lease                                    */
    /* This state raised by both clients: */
    DHC_TIMEDOUT,               /* libdhcp_control timeout has been exceeded          */
    /* DHCPv6 client states:    */
    DHC6_BOUND,                 /* new lease obtained             - arg is optinfo *  */
    DHC6_REBIND,                /* existing expired lease rebound - arg is optinfo *  */
    DHC6_RELEASE                /* existing lease expired         - arg is dhcp6_iaidaddr*/
} DHCP_State;

extern char *libdhcp_state_string( DHCP_State, char *buf );

/* default error handlers:
 */
extern 
int libdhcp_stderr_logger
(   struct libdhcp_control_s *ctl,
    int priority,  /* ala syslog(3): LOG_EMERG=0 - LOG_DEBUG=7 (+ LOG_FATAL=8 : finished -> 1)   */
    const char *fmt,
    va_list ap
);

extern 
int libdhcp_syslogger
(   struct libdhcp_control_s *ctl,
    int priority,  /* ala syslog(3): LOG_EMERG=0 - LOG_DEBUG=7 (+ LOG_FATAL=8 : finished -> 1)   */
    const char *fmt,
    va_list ap
);

#endif

